/* 2D potential flow solutions
   implemented with complex variables
   dudley.benton@gmail.com */

/* the first non-zero will select which case */
#define CIRCULATION     0
#define CORNER          0
#define CYLINDER        0
#define DOUBLET         0
#define HALF_BODY       0
#define NOZZLE          1
#define OVER_STEP       0
#define PLATE           0
#define RANKINE         0
#define SOURCE          0
#define SOURCE_SINK     0
#define STAGNATION      0
#define UNIFORM         0
#define VORTEX          0
#define WEDGE           0

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "../include/complex.hpp"

typedef struct{complex v,w;}vars;

vars UniformFlow(double angle,complex z)
  {
  static vars s;
  s.w=z*exp(-I*angle);
  s.v=exp(I*angle);
  return(s);
  }

vars PointSource(complex z)
  {
  double r;
  static vars s;
  s.w=s.v=complex(0.,0.);
  r=abs(z);
  if(r>DBL_EPSILON)
    {
    double pi2;
    pi2=M_PI*2.;
    s.w=log(z)/pi2;
    s.v=z/r/r/pi2;
    }
  return(s);
  }

vars PointDoublet(complex z)
  {
  double r;
  static vars s;
  s.w=s.v=complex(0.,0.);
  r=abs(z);
  if(r>DBL_EPSILON)
    {
    double rr,rrrr;
    rr=r*r;
    rrrr=rr*rr;
    s.w=conj(z)/rr;
    s.v.re=(z.im*z.im-z.re*z.re)/rrrr;
    s.v.im=-2.*z.re*z.im/rrrr;
    }
  return(s);
  }

vars CylindricalVortex(complex z)
  {
  double a,pi2,r;
  static vars s;
  s.w=s.v=complex(0.,0.);
  r=abs(z);
  if(r>DBL_EPSILON)
    {
    a=arg(z);
    pi2=M_PI*2.;
    s.w.re=-a/pi2;
    s.w.im=-log(r)/pi2;
    s.v.re=-z.im/r/r/pi2;
    s.v.im=z.re/r/r/pi2;
    }
  return(s);
  }

vars FlowPastHalfBody(double a,complex z)
  {
  static vars s,t;
  t=UniformFlow(0.,z);
  s=PointSource(z);
  t.w+=a*s.w;
  t.v+=a*s.v;
  return(t);
  }

vars FlowPastCylinder(double a,double b,complex z)
  {
  static vars s,t;
  t=UniformFlow(a,z);
  s=PointDoublet(z);
  t.w+=b*s.w;
  t.v+=b*s.v;
  return(t);
  }

vars FlowPastCylinderWithCirculation(double a,double b,complex z)
  {
  static vars s,t;
  t=UniformFlow(0.,z);
  s=PointDoublet(z);
  t.w+=a*s.w;
  t.v+=a*s.v;
  s=CylindricalVortex(z);
  t.w+=b*s.w;
  t.v+=b*s.v;
  return(t);
  }

vars FlowPastRankineBody(double a,double b,complex z)
  {
  static vars s,t;
  t=UniformFlow(0.,z);
  s=PointSource(z+b);
  t.w+=a*s.w;
  t.v+=a*s.v;
  s=PointSource(z-b);
  t.w-=a*s.w;
  t.v-=a*s.v;
  return(t);
  }

complex Plate(double angle,double L,complex z)
  {
  complex q;
  static vars s;
  if(z.re>=0.)
    q=z;
  else
    {
    q.re=fabs(z.re);
    q.im=-z.im;
    }
  s=UniformFlow(angle,(q+sqrt(q*q-L*L/4.))/2.);
  if(z.re<0.)
    return(-s.w);
  return(s.w);
  }

vars FlowOverPlate(double angle,double L,complex z)
  {
  double delta=0.0001;
  complex p1,p2;
  static vars s;
  p1=Plate(angle,L,z-delta/2.);
  p2=Plate(angle,L,z+delta/2.);
  s.w=(p1+p2)/2.;
  s.v.re=(p2.re-p1.re)/delta;
  s.v.im=(p1.im-p2.im)/delta;
  return(s);
  }

vars Power(double n,complex z)
  {
  double r,t;
  static vars s;
  s.w=s.v=complex(0.,0.);
  r=abs(z);
  if(r>DBL_EPSILON)
    {
    s.w=pow(z,n);
    t=arg(z);
    s.v.re= n*(z.re*cos(n*t)+z.im*sin(n*t))/pow(r,2.*n);
    s.v.im=-n*(z.re*sin(n*t)-z.im*cos(n*t))/pow(r,2.*n);
    }
  return(s);
  }

vars Nozzle(double d,complex z)
  {
  return(Power(1.,z*d-LambertW(exp(z*d))));
  }

int n=50;
int m=50;
double H=1.;
double D=1.;
double Sm,Sx;

void CreateV2D(char*fname)
  {
  int i,j;
  FILE*fp;
  complex z;
  static vars s1,s2,t;
  printf("writing velocities: %s\n",fname);
  fp=fopen(fname,"wt");
  for(j=0;j<m;j++)
    {
    z.im=H*(2*j+1-m)/(m-1);
    for(i=0;i<n;i++)
      {
      z.re=D*(2*i+1-n)/(n-1);
#if(CIRCULATION)
      t=FlowPastCylinderWithCirculation(D/6.,-3.*D,z);
      if(abs(z)<1./3.)
        continue;
#elif(CORNER)
      t=Power(2.,z);
#elif(CYLINDER)
      t=FlowPastCylinder(M_PI/4.,D/6.,z);
      if(abs(z)<1./3.)
        continue;
#elif(DOUBLET)
      t=PointDoublet(z);
#elif(HALF_BODY)
      t=FlowPastHalfBody(D,z);
#elif(NOZZLE)
      t=Nozzle(M_PI,z);
#elif(OVER_STEP)
      if(fabs(z.im)<z.re)
        continue;
      t=Power(2./3.,z);
#elif(PLATE)
      if(fabs(z.re)<0.5&&fabs(z.im)<0.05)
        continue;
      t=FlowOverPlate(M_PI/4.,D,z);
#elif(RANKINE)
      if(abs(z)<1./3.)
        continue;
      t=FlowPastRankineBody(D,D/4.,z);
#elif(SOURCE)
      t=PointSource(z);
#elif(SOURCE_SINK)
      s1=PointSource(z+0.5);
      s2=PointSource(z-0.5);
      t.w=s1.w-s2.w;
      t.v=s1.v-s2.v;
#elif(STAGNATION)
      s1=PointSource(z+1.);
      s2=PointSource(z-1.);
      t.w=-s1.w-s2.w;
      t.v=-s1.v-s2.v;
#elif(UNIFORM)
      t=UniformFlow(0.1,z);
#elif(VORTEX)
      t=CylindricalVortex(z);
#elif(WEDGE)
      t=Power(3.,z);
#else
      fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
      exit(1);
#endif
      if(abs(t.v)<10.)
        fprintf(fp,"%lG\t%lG\t%lG\t%lG\n",z.re,z.im,t.v.re,t.v.im);
      }
    }
  fclose(fp);
  }

void CreateTB2(char*fname,int which)
  {
  int i,j;
  complex z;
  FILE*fp;
  static vars s1,s2,t;
  printf("writing potential: %s\n",fname);
  fp=fopen(fname,"wt");
  fprintf(fp,"%i\n",n);
  for(i=0;i<n;i++)
    {
    z.re=D*(2*i+1-n)/(n-1);
    fprintf(fp,"%lG\n",z.re);
    }
  fprintf(fp,"%i\n",m);
  for(j=0;j<m;j++)
    {
    z.im=H*(2*j+1-m)/(m-1);
    fprintf(fp,"%lG\n",z.im);
    }
  fprintf(fp,"%i\n",n*m);
  Sm=DBL_MAX;
  Sx=-DBL_MAX;
  for(j=0;j<m;j++)
    {
    z.im=H*(2*j+1-m)/(m-1);
    for(i=0;i<n;i++)
      {
      z.re=D*(2*i+1-n)/(n-1);
#if(CIRCULATION)
      t=FlowPastCylinderWithCirculation(D/6.,-3.*D,z);
#elif(CORNER)
      t=Power(2.,z);
#elif(CYLINDER)
      t=FlowPastCylinder(M_PI/4.,D/6.,z);
#elif(DOUBLET)
      t=PointDoublet(z);
#elif(HALF_BODY)
      t=FlowPastHalfBody(D,z);
#elif(NOZZLE)
      t=Nozzle(M_PI,z);
#elif(OVER_STEP)
      t=Power(2./3.,z);
#elif(PLATE)
      t=FlowOverPlate(M_PI/4.,D,z);
#elif(RANKINE)
      t=FlowPastRankineBody(D,D/4.,z);
#elif(SOURCE)
      t=PointSource(z);
#elif(SOURCE_SINK)
      s1=PointSource(z+0.5);
      s2=PointSource(z-0.5);
      t.w=s1.w-s2.w;
      t.v=s1.v-s2.v;
#elif(STAGNATION)
      s1=PointSource(z+1.);
      s2=PointSource(z-1.);
      t.w=-s1.w-s2.w;
      t.v=-s1.v-s2.v;
#elif(UNIFORM)
      t=UniformFlow(M_PI/4.,z);
#elif(VORTEX)
      t=CylindricalVortex(z);
#elif(WEDGE)
      t=Power(3.,z);
#else
      fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
      exit(1);
#endif
      if(which%2)
        {
        fprintf(fp,"%lG\n",t.w.re);
        if(t.w.re<Sm)
          Sm=t.w.re;
        if(t.w.re>Sx)
          Sx=t.w.re;
        }
      else
        {
        fprintf(fp,"%lG\n",t.w.im);
        if(t.w.im<Sm)
          Sm=t.w.im;
        if(t.w.im>Sx)
          Sx=t.w.im;
        }
      }
    }
  fclose(fp);
  printf("%lG<%s<%lG\n",Sm,(which%2)?"psi":"phi",Sx);
  }

void CreateTP2(char*name,int which)
  {
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  fprintf(fp,"TP2\n");
  fprintf(fp,"!PLOT\n");
  fprintf(fp,"EQUAL\n");
  fprintf(fp,"TICKS\n");
  fprintf(fp,"LEGENDS\n");
  fprintf(fp,"-1 5 2\n");
  fprintf(fp,"CONTOUR\n");
  fprintf(fp,"%lG 0 %lG\n",Sm,Sx);
  fprintf(fp,"RAINBOW\n");
  fprintf(fp,"ZTABLE\n");
  fprintf(fp,"CENTER\n");
  fprintf(fp,"ABOVE\n");
#if(CIRCULATION)
  fprintf(fp,"Flow Past Cylinder with Circulation\n");
#elif(CORNER)
  fprintf(fp,"Four 90-degree Corners\n");
#elif(CYLINDER)
  fprintf(fp,"Flow Past Cylinder\n");
#elif(DOUBLET)
  fprintf(fp,"Doublet\n");
#elif(HALF_BODY)
  fprintf(fp,"Flow Past a Half-Body\n");
#elif(NOZZLE)
  fprintf(fp,"Flow from a Nozzle\n");
#elif(OVER_STEP)
  fprintf(fp,"Flow Over a Step\n");
#elif(PLATE)
  fprintf(fp,"Plate in an Inclined Flow\n");
#elif(RANKINE)
  fprintf(fp,"Flow Past Rankine Body\n");
#elif(SOURCE)
  fprintf(fp,"Source\n");
#elif(SOURCE_SINK)
  fprintf(fp,"Source and Sink\n");
#elif(STAGNATION)
  fprintf(fp,"Stagnation Flow\n");
#elif(UNIFORM)
  fprintf(fp,"Inclined Uniform Flow\n");
#elif(VORTEX)
  fprintf(fp,"Cylindrical (Free) Vortex\n");
#elif(WEDGE)
  fprintf(fp,"Four Wedges\n");
#else
  fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
  exit(1);
#endif
  fprintf(fp,"X\n");
  fprintf(fp,"%lG %lG\n",-D,D);
  fprintf(fp,"Y\n");
  fprintf(fp,"%lG %lG\n",-H,H);
  if(which%2)
    fprintf(fp,"cmplxflo1.tb2\n");
  else
    fprintf(fp,"cmplxflo2.tb2\n");
  fprintf(fp,"ARROW\n");
  fprintf(fp,"0 0\n");
  fprintf(fp,"NEWFILE\n");
  fprintf(fp,"cmplxflo.v2d\n");
  fprintf(fp,"NOCOLOR\n");
  fprintf(fp,"!END\n");
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  CreateV2D("cmplxflo.v2d");
  CreateTB2("cmplxflo1.tb2",1);
  CreateTP2("cmplxflo1.tp2",1);
  CreateTB2("cmplxflo2.tb2",2);
  CreateTP2("cmplxflo2.tp2",2);
  return(0);
  }
