/* 2D potential flow solutions
   implemented using real (non-complex) variables
   dudley.benton@gmail.com */

#define CIRCULATION 0
#define CYLINDER    0
#define DOUBLET     0
#define HALF_BODY   0
#define RANKINE     1
#define SOURCE      0
#define SOURCE_SINK 0
#define STAGNATION  0
#define UNIFORM     0

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

typedef struct{double phi,psi,r,theta,u,v,x,y;}vars;

vars CartesianToCylindrical(double x,double y)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s.x=x;
  s.y=y;
  s.r=hypot(x,y);
  if(s.r>DBL_EPSILON)
    s.theta=atan2(y,x);
  return(s);
  }

vars UniformFlow(double angle,double x,double y)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToCylindrical(x,y);
  s.u=cos(angle);
  s.v=sin(angle);
  s.psi=x*s.u+y*s.v;
  s.phi=y*s.u-x*s.v;
  return(s);
  }

vars PointSource(double x,double y)
  {
  double pi2;
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToCylindrical(x,y);
  if(s.r>DBL_EPSILON)
    {
    pi2=M_PI*2.;
    s.phi=log(s.r)/pi2;
    s.psi=s.theta/pi2;
    s.u=x/pow(s.r,2)/pi2;
    s.v=y/pow(s.r,2)/pi2;
    }
  return(s);
  }

vars PointDoublet(double x,double y)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToCylindrical(x,y);
  if(s.r>DBL_EPSILON)
    {
    s.phi=x/pow(s.r,2);
    s.psi=-y/pow(s.r,2);
    s.u=(pow(y,2)-pow(x,2))/pow(s.r,4);
    s.v=-2.*x*y/pow(s.r,4);
    }
  return(s);
  }

vars CylindricalVortex(double x,double y)
  {
  double pi2;
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToCylindrical(x,y);
  if(s.r>DBL_EPSILON)
    {
    pi2=M_PI*2.;
    s.phi=-s.theta/pi2;
    s.psi=-log(s.r)/pi2;
    s.u=-y/pow(s.r,2)/pi2;
    s.v=x/pow(s.r,2)/pi2;
    }
  return(s);
  }

vars FlowPastHalfBody(double a,double x,double y)
  {
  static vars s,t;
  t=UniformFlow(0.,x,y);
  s=PointSource(x,y);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  return(t);
  }

vars FlowPastCylinder(double a,double x,double y)
  {
  static vars s,t;
  t=UniformFlow(0.,x,y);
  s=PointDoublet(x,y);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  return(t);
  }

vars FlowPastCylinderWithCirculation(double a,double b,double x,double y)
  {
  static vars s,t;
  t=UniformFlow(0.,x,y);
  s=PointDoublet(x,y);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  s=CylindricalVortex(x,y);
  t.phi+=b*s.phi;
  t.psi+=b*s.psi;
  t.u+=b*s.u;
  t.v+=b*s.v;
  return(t);
  }

vars FlowPastRankineBody(double a,double b,double x,double y)
  {
  static vars s,t;
  t=UniformFlow(0.,x,y);
  s=PointSource(x+b,y);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  s=PointSource(x-b,y);
  t.phi-=a*s.phi;
  t.psi-=a*s.psi;
  t.u-=a*s.u;
  t.v-=a*s.v;
  return(t);
  }

int n=50;
int m=50;
double H=1.;
double D=1.;
double Sm,Sx;

void CreateV2D(char*fname)
  {
  int i,j;
  double x,y;
  FILE*fp;
  static vars s1,s2,t;
  printf("writing velocities: %s\n",fname);
  fp=fopen(fname,"wt");
  for(j=0;j<m;j++)
    {
    y=H*(2*j+1-m)/(m-1);
    for(i=0;i<n;i++)
      {
      x=D*(2*i+1-n)/(n-1);
#if(CIRCULATION)
      t=FlowPastCylinderWithCirculation(D/6.,-3.*D,x,y);
      if(t.r<1./3.)
        continue;
#elif(CYLINDER)
      t=FlowPastCylinder(D/6.,x,y);
      if(t.r<1./3.)
        continue;
#elif(DOUBLET)
      t=PointDoublet(x,y);
#elif(HALF_BODY)
      t=FlowPastHalfBody(D,x,y);
#elif(RANKINE)
      if(fabs(x)<1./3.&&fabs(y)<1./4.)
        continue;
      t=FlowPastRankineBody(D,D/4.,x,y);
#elif(SOURCE)
      t=PointSource(x,y);
#elif(SOURCE_SINK)
      s1=PointSource(x+0.5,y);
      s2=PointSource(x-0.5,y);
      t.psi=s1.psi-s2.psi;
      t.phi=s1.phi-s2.phi;
      t.u=s1.u-s2.u;
      t.v=s1.v-s2.v;
#elif(STAGNATION)
      s1=PointSource(x+1.,y);
      s2=PointSource(x-1.,y);
      t.psi=-s1.psi-s2.psi;
      t.phi=-s1.phi-s2.phi;
      t.u=-s1.u-s2.u;
      t.v=-s1.v-s2.v;
#elif(UNIFORM)
      t=UniformFlow(M_PI/4.,x,y);
#else
      fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
      exit(1);
#endif
      if(hypot(t.u,t.v)<10.)
        fprintf(fp,"%lG\t%lG\t%lG\t%lG\n",x,y,t.u,t.v);
      }
    }
  fclose(fp);
  }

void CreateTB2(char*fname,int which)
  {
  int i,j;
  double x,y;
  FILE*fp;
  static vars s1,s2,t;
  printf("writing potential: %s\n",fname);
  fp=fopen(fname,"wt");
  fprintf(fp,"%i\n",n);
  for(i=0;i<n;i++)
    {
    x=D*(2*i+1-n)/(n-1);
    fprintf(fp,"%lG\n",x);
    }
  fprintf(fp,"%i\n",m);
  for(j=0;j<m;j++)
    {
    y=H*(2*j+1-m)/(m-1);
    fprintf(fp,"%lG\n",y);
    }
  fprintf(fp,"%i\n",n*m);
  Sm=DBL_MAX;
  Sx=-DBL_MAX;
  for(j=0;j<m;j++)
    {
    y=H*(2*j+1-m)/(m-1);
    for(i=0;i<n;i++)
      {
      x=D*(2*i+1-n)/(n-1);
#if(CIRCULATION)
      t=FlowPastCylinderWithCirculation(D/6.,-3.*D,x,y);
#elif(CYLINDER)
      t=FlowPastCylinder(D/6.,x,y);
#elif(DOUBLET)
      t=PointDoublet(x,y);
#elif(HALF_BODY)
      t=FlowPastHalfBody(D,x,y);
#elif(RANKINE)
      t=FlowPastRankineBody(D,D/4.,x,y);
#elif(SOURCE)
      t=PointSource(x,y);
#elif(SOURCE_SINK)
      s1=PointSource(x+0.5,y);
      s2=PointSource(x-0.5,y);
      t.psi=s1.psi-s2.psi;
      t.phi=s1.phi-s2.phi;
      t.u=s1.u-s2.u;
      t.v=s1.v-s2.v;
#elif(STAGNATION)
      s1=PointSource(x+1.,y);
      s2=PointSource(x-1.,y);
      t.psi=-s1.psi-s2.psi;
      t.phi=-s1.phi-s2.phi;
      t.u=-s1.u-s2.u;
      t.v=-s1.v-s2.v;
#elif(UNIFORM)
      t=UniformFlow(M_PI/4.,x,y);
#else
      fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
      exit(1);
#endif
      if(which%2)
        {
        fprintf(fp,"%lG\n",t.phi);
        if(t.phi<Sm)
          Sm=t.phi;
        if(t.phi>Sx)
          Sx=t.phi;
        }
      else
        {
        fprintf(fp,"%lG\n",t.psi);
        if(t.psi<Sm)
          Sm=t.psi;
        if(t.psi>Sx)
          Sx=t.psi;
        }
      }
    }
  fclose(fp);
  printf("%lG<%s<%lG\n",Sm,(which%2)?"psi":"phi",Sx);
  }

void CreateTP2(char*name,int which)
  {
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  fprintf(fp,"TP2\n");
  fprintf(fp,"!PLOT\n");
  fprintf(fp,"EQUAL\n");
  fprintf(fp,"TICKS\n");
  fprintf(fp,"LEGENDS\n");
  fprintf(fp,"-1 5 2\n");
  fprintf(fp,"CONTOUR\n");
  fprintf(fp,"%lG 0 %lG\n",Sm,Sx);
  fprintf(fp,"RAINBOW\n");
  fprintf(fp,"ZTABLE\n");
  fprintf(fp,"CENTER\n");
  fprintf(fp,"ABOVE\n");
#if(CIRCULATION)
  fprintf(fp,"Flow Past Cylinder with Circulation\n");
#elif(CYLINDER)
  fprintf(fp,"Flow Past Cylinder\n");
#elif(DOUBLET)
  fprintf(fp,"Doublet\n");
#elif(HALF_BODY)
  fprintf(fp,"Flow Past a Half-Body\n");
#elif(RANKINE)
  fprintf(fp,"Flow Past Rankine Body\n");
#elif(SOURCE)
  fprintf(fp,"Source\n");
#elif(SOURCE_SINK)
  fprintf(fp,"Source and Sink\n");
#elif(STAGNATION)
  fprintf(fp,"Stagnation Flow\n");
#elif(UNIFORM)
  fprintf(fp,"Inclined Uniform Flow\n");
#else
  fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
  exit(1);
#endif
  fprintf(fp,"X\n");
  fprintf(fp,"%lG %lG\n",-D,D);
  fprintf(fp,"Y\n");
  fprintf(fp,"%lG %lG\n",-H,H);
  if(which%2)
    fprintf(fp,"potflow1.tb2\n");
  else
    fprintf(fp,"potflow2.tb2\n");
  fprintf(fp,"ARROW\n");
  fprintf(fp,"0 0\n");
  fprintf(fp,"NEWFILE\n");
  fprintf(fp,"potflow.v2d\n");
  fprintf(fp,"NOCOLOR\n");
  fprintf(fp,"!END\n");
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  CreateV2D("potflow.v2d");
  CreateTB2("potflow1.tb2",1);
  CreateTP2("potflow1.tp2",1);
  CreateTB2("potflow2.tb2",2);
  CreateTP2("potflow2.tp2",2);
  return(0);
  }
