/* 3D potential flow solutions
   implemented using real (non-complex) variables
   dudley.benton@gmail.com */

#define DOUBLET     0
#define HALF_BODY   0
#define RANKINE     0
#define SOURCE      0
#define SOURCE_SINK 0
#define SPHERE      0
#define STAGNATION  1

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

double hypot3(double x,double y,double z)
  {
  return(sqrt(x*x+y*y+z*z));
  }

typedef struct{double alpha,phi,psi,r,s,theta,u,v,w,x,y,z;}vars;

vars CartesianToSpherical(double x,double y,double z)
  {
  static vars s;
  s.x=x;
  s.y=y;
  s.z=z;
  s.s=hypot(x,y);
  s.r=hypot3(x,y,z);
  if(s.r>DBL_EPSILON)
    s.alpha=atan2(z,s.s);
  if(s.s>DBL_EPSILON)
    s.theta=atan2(y,x);
  return(s);
  }

vars UniformFlow(double x,double y,double z)
  {
  static vars s;
  s=CartesianToSpherical(x,y,z);
  s.phi=-z;
  s.psi=(x*x+y*y)/2.;
  s.w=1.;
  return(s);
  }

vars PointSource(double x,double y,double z)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToSpherical(x,y,z);
  if(s.r>DBL_EPSILON)
    {
    s.phi=-1./s.r;
    s.psi=-z/s.r;
    s.u=x/pow(s.r,3);
    s.v=y/pow(s.r,3);
    s.w=z/pow(s.r,3);
    }
  return(s);
  }

vars PointDoublet(double x,double y,double z)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToSpherical(x,y,z);
  if(s.r>DBL_EPSILON)
    {
    s.phi=       z      /pow(s.r,3);
    s.psi=  -(x*x+y*y)  /pow(s.r,3);
    s.u=   3.*x*z       /pow(s.r,5);
    s.v=   3.*y*z       /pow(s.r,5);
    s.w=(2.*z*z-x*x-y*y)/pow(s.r,5);
    }
  return(s);
  }

vars LineSource(double x,double y,double z)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToSpherical(x,y,z);
  if(s.s>DBL_EPSILON)
    {
    s.phi=-1./s.s;
    s.psi=-z;
    s.u=x/pow(s.s,3);
    s.v=y/pow(s.s,3);
    s.w=0.;
    }
  return(s);
  }

vars CylindricalVortex(double q,double x,double y,double z)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToSpherical(x,y,z);
  if(s.s>DBL_EPSILON)
    {
    s.phi=s.alpha;
    s.psi=-log(s.s/q);
    s.u=-y;
    s.v=x;
    s.w=0.;
    }
  return(s);
  }

vars SphericalVortex(double q,double x,double y,double z)
  {
  static vars s;
  memset(&s,0,sizeof(s));
  s=CartesianToSpherical(x,y,z);
  if(s.r>DBL_EPSILON)
    {
    s.phi=(x*x+y*y)*(2+pow(q,3)/pow(s.r,3));
    s.psi=   z*z   *(1-pow(q,3)/pow(s.r,3));
    if(s.s>DBL_EPSILON)
      {
      s.u=((3.*s.s/pow(s.r,2)-1./s.s)*pow(q,3)/pow(s.r,3)-2./s.s)*x;
      s.v=((3.*s.s/pow(s.r,2)-1./s.s)*pow(q,3)/pow(s.r,3)-2./s.s)*y;
      s.w=  3.*s.s/pow(s.r,2)        *pow(q,3)/pow(s.r,3)        *z;
      }
    }
  return(s);
  }

vars FlowPastHalfBody(double a,double x,double y,double z)
  {
  static vars s,t;
  t=UniformFlow(x,y,z);
  s=PointSource(x,y,z);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  t.w+=a*s.w;
  return(t);
  }

vars FlowPastSphere(double a,double x,double y,double z)
  {
  static vars s,t;
  t=UniformFlow(x,y,z);
  s=PointDoublet(x,y,-z);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  t.w+=a*s.w;
  return(t);
  }

vars FlowPastSphereWithCirculation(double a,double b,double x,double y,double z)
  {
  static vars s,t;
  t=UniformFlow(x,y,z);
  s=PointDoublet(x,y,-z);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  t.w+=a*s.w;
  s=CylindricalVortex(b,x,y,z);
  t.phi+=s.phi;
  t.psi+=s.psi;
  t.u+=s.u;
  t.v+=s.v;
  t.w+=s.w;
  return(t);
  }

vars FlowPastRankineBody(double a,double b,double x,double y,double z)
  {
  static vars s,t;
  t=UniformFlow(x,y,z);
  s=PointSource(x,y,z+b);
  t.phi+=a*s.phi;
  t.psi+=a*s.psi;
  t.u+=a*s.u;
  t.v+=a*s.v;
  t.w+=a*s.w;
  s=PointSource(x,y,z-b);
  t.phi-=a*s.phi;
  t.psi-=a*s.psi;
  t.u-=a*s.u;
  t.v-=a*s.v;
  t.w-=a*s.w;
  return(t);
  }

int l=10;
int m=10;
int n=10;
double W=1.; /* X-dimension (width) */
double H=1.; /* Y-dimension (height) */
double D=1.; /* Z-dimension (depth) */
double Sm,Sx;

void CreateV3D(char*fname)
  {
  int i,j,k;
  double x,y,z;
  FILE*fp;
  static vars s1,s2,t;
  printf("writing velocities: %s\n",fname);
  fp=fopen(fname,"wt");
  for(k=0;k<n;k++)
    {
    z=D*(2*k+1-n)/(n-1);
    for(j=0;j<m;j++)
      {
      y=H*(2*j+1-m)/(m-1);
      for(i=0;i<l;i++)
        {
        x=W*(2*i+1-l)/(l-1);
#if(DOUBLET)
        t=PointDoublet(x,y,z);
#elif(HALF_BODY)
        t=FlowPastHalfBody(D,x,y,z);
#elif(RANKINE)
        t=FlowPastRankineBody(D/2.,D/2.,x,y,z);
#elif(SOURCE)
        t=PointSource(x,y,z);
#elif(SOURCE_SINK)
        s1=PointSource(x,y+0.5,z);
        s2=PointSource(x,y-0.5,z);
        t.phi=s1.phi-s2.phi;
        t.psi=s1.psi-s2.psi;
        t.u=s1.u-s2.u;
        t.v=s1.v-s2.v;
        t.w=s1.w-s2.w;
#elif(SPHERE)
        t=FlowPastSphereWithCirculation(D,D,x,y,z);
#elif(STAGNATION)
        s1=PointSource(x,y,z+1.);
        s2=PointSource(x,y,z-1.);
        t.phi=-s1.phi-s2.phi;
        t.psi=-s1.psi-s2.psi;
        t.u=-s1.u-s2.u;
        t.v=-s1.v-s2.v;
        t.w=-s1.w-s2.w;
#else
        fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
        exit(1);
#endif
        if(hypot3(t.u,t.v,t.w)<10.)
          fprintf(fp,"%lG\t%lG\t%lG\t%lG\t%lG\t%lG\n",x,y,z,t.u,t.v,t.w);
        }
      }
    }
  fclose(fp);
  }

void CreateTB3(char*fname,int which)
  {
  int i,j,k;
  double x,y,z;
  FILE*fp;
  static vars s1,s2,t;
  printf("writing potential: %s\n",fname);
  fp=fopen(fname,"wt");
  fprintf(fp,"%i\n",l);
  for(i=0;i<l;i++)
    {
    x=W*(2*i+1-l)/(l-1);
    fprintf(fp,"%lG\n",x);
    }
  fprintf(fp,"%i\n",m);
  for(j=0;j<m;j++)
    {
    y=H*(2*j+1-m)/(m-1);
    fprintf(fp,"%lG\n",y);
    }
  fprintf(fp,"%i\n",n);
  for(k=0;k<n;k++)
    {
    z=D*(2*k+1-n)/(n-1);
    fprintf(fp,"%lG\n",z);
    }
  Sm=DBL_MAX;
  Sx=-DBL_MAX;
  fprintf(fp,"%i\n",l*m*n);
  for(k=0;k<n;k++)
    {
    z=D*(2*k+1-n)/(n-1);
    for(j=0;j<m;j++)
      {
      y=H*(2*j+1-m)/(m-1);
      for(i=0;i<l;i++)
        {
        x=W*(2*i+1-l)/(l-1);
#if(DOUBLET)
        t=PointDoublet(x,y,z);
#elif(HALF_BODY)
        t=FlowPastHalfBody(D,x,y,z);
#elif(RANKINE)
        t=FlowPastRankineBody(D/2.,D/2.,x,y,z);
#elif(SOURCE)
        t=PointSource(x,y,z);
#elif(SOURCE_SINK)
        s1=PointSource(x,y+0.5,z);
        s2=PointSource(x,y-0.5,z);
        t.phi=s1.phi-s2.phi;
        t.psi=s1.psi-s2.psi;
        t.u=s1.u-s2.u;
        t.v=s1.v-s2.v;
        t.w=s1.w-s2.w;
#elif(SPHERE)
        t=FlowPastSphereWithCirculation(D,D,x,y,z);
#elif(STAGNATION)
        s1=PointSource(x,y,z+1.);
        s2=PointSource(x,y,z-1.);
        t.phi=-s1.phi-s2.phi;
        t.psi=-s1.psi-s2.psi;
        t.u=-s1.u-s2.u;
        t.v=-s1.v-s2.v;
        t.w=-s1.w-s2.w;
#else
        fprintf(stderr,"unexpected compiler option at line %i\n",__LINE__);
        exit(1);
#endif
        if(which%2)
          {
          fprintf(fp,"%lG\n",t.psi);
          if(t.psi<Sm)
            Sm=t.psi;
          if(t.psi>Sx)
            Sx=t.psi;
          }
        else
          {
          fprintf(fp,"%lG\n",t.phi);
          if(t.phi<Sm)
            Sm=t.phi;
          if(t.phi>Sx)
            Sx=t.phi;
          }
        }
      }
    }
  fclose(fp);
  printf("%lG<%s<%lG\n",Sm,(which%2)?"psi":"phi",Sx);
  }

void CreateTP2(char*name)
  {
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  fprintf(fp,"TP2\n");
  fprintf(fp,"!PLOT\n");
  fprintf(fp,"EQUAL\n");
  fprintf(fp,"TICKS\n");
  fprintf(fp,"LEGENDS\n");
  fprintf(fp,"-1 5 2\n");
  fprintf(fp,"CONTOUR\n");
  fprintf(fp,"%lG 0 %lG\n",Sm,Sx);
  fprintf(fp,"RAINBOW\n");
  fprintf(fp,"ZTABLE\n");
  fprintf(fp,"CENTER\n");
  fprintf(fp,"ABOVE\n");
#if(DOUBLET)
  fprintf(fp,"Doublet\n");
#elif(HALF_BODY)
  fprintf(fp,"Flow Past a Half-Body\n");
#elif(RANKINE)
  fprintf(fp,"Flow Past Rankine Body\n");
#elif(SOURCE)
  fprintf(fp,"Source\n");
#elif(SOURCE_SINK)
  fprintf(fp,"Source and Sink\n");
#elif(SPHERE)
  fprintf(fp,"Flow Past Sphere with Circulation\n");
#elif(STAGNATION)
  fprintf(fp,"Stagnation Flow\n");
#else
  exit(1);
#endif
  fprintf(fp,"X\n");
  fprintf(fp,"%lG %lG\n",-D,D);
  fprintf(fp,"Y\n");
  fprintf(fp,"%lG %lG\n",-H,H);
  fprintf(fp,"potflow1.tb3\n");
  fprintf(fp,"ARROW\n");
  fprintf(fp,"0 0\n");
  fprintf(fp,"NEWFILE\n");
  fprintf(fp,"potflow.v2d\n");
  fprintf(fp,"NOCOLOR\n");
  fprintf(fp,"!END\n");
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  CreateV3D("potflow.v3d");
  CreateTB3("potflow1.tb3",1);
  CreateTB3("potflow2.tb3",2);
  CreateTP2("potflow.tp2");
  return(0);
  }
