/* Bairstow's method for real polynomial roots
   (some of which may be complex)
   developed by Dudley J. Benton
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

typedef struct{
  double re;
  double im;
  }COMPLEX;

COMPLEX Complex(double re,double im)
  {
  COMPLEX c;
  c.re=re;
  c.im=im;
  return(c);
  }

void QuadratricRoots(double*c,COMPLEX*r)
  {
  double d;
  r[0]=r[1]=Complex(0.,0.);
  d=c[0]*c[0]-4.*c[1];
  if(d>=0.)
    {
    d=sqrt(d);
    r[0]=Complex((-c[0]+d)/2.,0);
    r[1]=Complex((-c[0]-d)/2.,0);
    }
  else
    {
    d=sqrt(fabs(d));
    r[0]=Complex(-c[0]/2., d/2.);
    r[1]=Complex(-c[0]/2.,-d/2.);
    }
  }

void CubicRoots(double*c,COMPLEX*r)
  {
  double a,arm,b,ca,ca3,cb,cb3,d,rm,t,x;
  r[0]=r[1]=r[2]=Complex(0.,0.);
  a=c[1]-c[0]*c[0]/3.;
  b=(2.*c[0]*c[0]*c[0]-9.*c[0]*c[1]+27.*c[2])/27.;
  d=b*b/4.+a*a*a/27.;
  if(d>=0.)
    {
    d=sqrt(d);
    ca3=-b/2.+d;
    cb3=-b/2.-d;
    if(ca3<0.)
      ca=-pow(-ca3,1./3.);
    else if(ca3>0.)
      ca=pow(ca3,1./3.);
    else
      ca=0.;
    if(cb3<0.)
      cb=-pow(-cb3,1./3.);
    else if(cb3>0.)
      cb=pow(cb3,1./3.);
    else
      cb=0.;
    r[0]=Complex(ca+cb-c[0]/3.,0.);
    r[1]=Complex(-(ca+cb)/2.-c[0]/3., (ca-cb)*sqrt(3.)/2.);
    r[2]=Complex(-(ca+cb)/2.-c[0]/3.,-(ca-cb)*sqrt(3.)/2.);
    }
  else
    {
    rm=2.*sqrt(-a/3.);
    arm=a*rm;
    if(fabs(arm)>0.)
      x=3.*b/arm;
    else
      x=0.;
    t=acos(x)/3.;
    r[0]=Complex(rm*cos(t)-c[0]/3.,0.);
    r[1]=Complex(rm*cos(t+2.*M_PI/3.)-c[0]/3.,0.);
    r[2]=Complex(rm*cos(t+4.*M_PI/3.)-c[0]/3.,0.);
    }
  }

int Bairstow(double*a,int n,double*p,double*q) /* Bairstow's Method for roots */
  {
  int error=1,i,iter;
  double*b,*c,cpnm1,d,deltap,deltaq,dp,dq;
  b=calloc(n,sizeof(double));
  c=calloc(n,sizeof(double));
  for(iter=0;iter<100;iter++)
    {
    b[0]=a[0]-*p;
    c[0]=b[0]-*p;
    b[1]=a[1]-*p*b[0]-*q;
    c[1]=b[1]-*p*c[0]-*q;
    for(i=2;i<n;i++)
      {
      b[i]=a[i]-*p*b[i-1]-*q*b[i-2];
      c[i]=b[i]-*p*c[i-1]-*q*c[i-2];
      }
    cpnm1=c[n-2]-b[n-2];
    d=c[n-3]*c[n-3]-cpnm1*c[n-4];
    if(fabs(d)<FLT_MIN)
      {
      error=0;
      break;
      }
    dp=b[n-2]*c[n-3]-b[n-1]*c[n-4];
    dq=-b[n-2]*cpnm1+b[n-1]*c[n-3];
    deltap=dp/d;
    deltaq=dq/d;
    *p+=deltap;
    *q+=deltaq;
    if(fabs(deltap)+fabs(deltaq)<FLT_EPSILON)
      {
      error=0;
      break;
      }
    }
  if(error==0)
    for(i=0;i<(n-2);i++)
      a[i]=b[i];
  free(b);
  free(c);
  return(error);
  }

void PolynomialRoots(double*a,COMPLEX*r,int n)
  {
  int error,i;
  double*b,*c,p,q;
  for(i=0;i<n;i++)
    r[i]=Complex(0.,0.);
  if(fabs(a[n])<FLT_MIN)
    return;
  b=calloc(n,sizeof(double));
  c=calloc(n,sizeof(double));
  for(i=0;i<n;i++)
    c[i]=a[n-i-1]/a[n];
  while(n>3)
    {
    error=1;
    while(error)
      {
      i=rand()%n;
      p=a[i]*rand()/32767.;
      if(rand()%2)
        p=-p;
      i=rand()%n;
      q=a[i]*rand()/32767.;
      if(rand()%2)
        q=-q;
      error=Bairstow(c,n,&p,&q);
      }
    b[0]=p;
    b[1]=q;
    QuadratricRoots(b,&r[n-2]);
    n-=2;
    }
  if(n==3)
    CubicRoots(c,r);
  else if(n==2)
    QuadratricRoots(c,r);
  else if(n==1)
    r[0]=Complex(-c[0],0);
  free(b);
  free(c);
  }

int randbetween(int lo,int hi)
  {
  return(lo+rand()%(hi-lo+1));
  }

int main(int argc,char**argv,char**envp)
  {
  int i,j,n;
  double*a;
  COMPLEX*r;
  for(j=0;j<8;j++)
    {
    n=j<5?j+1:randbetween(1,9);
    a=calloc(n+1,sizeof(double));
    r=calloc(n+1,sizeof(COMPLEX));
    for(i=0;i<=n;i++)
      a[i]=j<5?i+1:randbetween(-10000,10000)/1000.;
    printf("a");
    for(i=0;i<=n;i++)
      printf("%c%lG%",i?',':'=',a[i]);
    printf("\n");
    PolynomialRoots(a,r,n);
    printf("r");
    for(i=0;i<n;i++)
      printf("%c(%lG%+lGi)",i?',':'=',r[i].re,r[i].im);
    printf("\n");
    }
  return(0);
  }
