/* 4th order Runge-Kutta for solving complex ODE
   developed by Dudley J. Benton
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "../include/complex.hpp"

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

double Sm,Sx;

void CreateTB2(complex*Z,complex*W,int m,int n,char*rname,char*iname)
  {
  int i,j;
  FILE*fr,*fi;
  printf("creating: %s\n",rname);
  fr=fopen(rname,"wt");
  printf("creating: %s\n",iname);
  fi=fopen(iname,"wt");
  fprintf(fr,"%i\n",n);
  fprintf(fi,"%i\n",n);
  for(j=0;j<n;j++)
    {
    fprintf(fr,"%lG\n",Z[j].re);
    fprintf(fi,"%lG\n",Z[j].re);
    }
  fprintf(fr,"%i\n",m);
  fprintf(fi,"%i\n",m);
  for(i=0;i<m;i++)
    {
    fprintf(fr,"%lG\n",Z[n*i].im);
    fprintf(fi,"%lG\n",Z[n*i].im);
    }
  Sm=DBL_MAX;
  Sx=-DBL_MAX;
  fprintf(fr,"%i\n",m*n);
  fprintf(fi,"%i\n",m*n);
  for(i=0;i<m;i++)
    {
    for(j=0;j<n;j++)
      {
      Sm=fmin(Sm,fmin(W[n*i+j].re,W[n*i+j].im));
      Sx=fmax(Sx,fmin(W[n*i+j].re,W[n*i+j].im));
      fprintf(fr,"%lG\n",W[n*i+j].re);
      fprintf(fi,"%lG\n",W[n*i+j].im);
      }
    }
  fclose(fr);
  fclose(fi);
  }

void CreateTP2(char*title,char*name,char*rname,char*iname)
  {
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  fprintf(fp,"TP2\n");
  fprintf(fp,"!ALTER\n");
  fprintf(fp,"!PLOT\n");
  fprintf(fp,"EQUAL\n");
  fprintf(fp,"TICKS\n");
  fprintf(fp,"LEGENDS\n");
  fprintf(fp,"-1 5 2\n");
  fprintf(fp,"RAINBOW\n");
  fprintf(fp,"CONTOUR\n");
  fprintf(fp,"%lG 0 %lG\n",Sm,Sx);
  fprintf(fp,"NOLEGEND 126\n");
  fprintf(fp,"ZTABLE\n");
  fprintf(fp,"CENTER\n");
  fprintf(fp,"ABOVE\n");
  fprintf(fp,"%s\n",title);
  fprintf(fp,"X\n");
  fprintf(fp,"-3 3\n");
  fprintf(fp,"Y\n");
  fprintf(fp,"-3 3\n");
  fprintf(fp,"%s\n",rname);
  fprintf(fp,"NEWFILE 2\n");
  fprintf(fp,"%s\n",iname);
  fprintf(fp,"!END\n");
  fclose(fp);
  }

void RungeKutta4(void dYdX(complex,complex*,complex*),complex X,complex dX,complex*Y,complex*dY,int n)
  {
  int i,j;
  double A[3]={0.5,0.5,1.};
  double B[4]={1./6.,1./3.,1./3.,1./6.};
  complex*V,*W;
  V=(complex*)calloc(5*n,sizeof(complex));
  W=V+n;
  dYdX(X,Y,W);
  for(i=1;i<4;i++)
    {
    for(j=0;j<n;j++)
      {
      dY[j]=A[i-1]*W[n*(i-1)+j];
      V[j]=Y[j]+dX*dY[j];
      }
    dYdX(X+dX*A[i-1],V,W+n*i);
    }
  for(j=0;j<n;j++)
    {
    dY[j]=0.;
    for(i=0;i<4;i++)
      dY[j]+=B[i]*W[n*i+j];
    Y[j]+=dX*dY[j];
    }
  free(V);
  }

void ODE(complex X,complex*Y,complex*dY)
  {
  dY[2]=-Y[0]*Y[2]/2.;
  dY[1]=Y[2];
  dY[0]=Y[1];
  }

#define m 64 /* number of points along y-axis */
#define n 64 /* number of points along x-axis */

complex W[m*n],dW[m*n],ddW[m*n],Z[m*n];

void SolveODE()
  {
  int i,j;
  complex dX,dY[3],X,Y[3];

/* initial conditions: X=(-3,-3) Y=1, dY/dX=0.1, d2Y/dX2=-0.01 */

  X.re=X.im=-3.;
  Y[0].re=Y[0].im=1.;
  Y[1].re=Y[1].im=0.1;
  Y[2].re=Y[2].im=-0.01;

/* step along x-axis (at lowest y-value) */

  Z[0]=X;
  W[0]=Y[0];
  dW[0]=Y[1];
  ddW[0]=Y[2];
  dX.re=6./(n-1);
  dX.im=0.;
  for(j=1;j<n;j++)
    {
    RungeKutta4(ODE,X,dX,Y,dY,3);
    X+=dX;
    Z[j]=X;
    W[j]=Y[0];
    dW[j]=Y[1];
    ddW[j]=Y[2];
    }

/* step along y-axis (one row of x-values at a time) */

  dX.re=0.;
  dX.im=6./(m-1);
  for(i=1;i<m;i++)
    {
    for(j=0;j<n;j++)
      {
      X=Z[n*(i-1)+j];
      Y[0]=W[n*(i-1)+j];
      Y[1]=dW[n*(i-1)+j];
      Y[2]=ddW[n*(i-1)+j];
      RungeKutta4(ODE,X,dX,Y,dY,3);
      Z[n*i+j]=X+dX;
      W[n*i+j]=Y[0];
      dW[n*i+j]=Y[1];
      ddW[n*i+j]=Y[2];
      }
    }
  }

void ListSolution()
  {
  int i,j;
  for(i=0;i<m;i++)
    {
    for(j=0;j<n;j++)
      {
      printf("(%lG%+lGi)",Z[n*i+j].re,Z[n*i+j].im);
      printf(",(%lG%+lGi)",W[n*i+j].re,W[n*i+j].im);
      printf(",(%lG%+lGi)",dW[n*i+j].re,dW[n*i+j].im);
      printf(",(%lG%+lGi)\n",ddW[n*i+j].re,ddW[n*i+j].im);
      }
    }
  }

int main(int argc,char**argv,char**envp)
  {
  SolveODE();
  CreateTB2(Z,W,n,m,"Yr.tb2","Yi.tb2");
  CreateTP2("W(z)","Y.tp2","Yr.tb2","Yi.tb2");
  CreateTB2(Z,dW,n,m,"dYr.tb2","dYi.tb2");
  CreateTP2("dW(z)/dz","dY.tp2","dYr.tb2","dYi.tb2");
  CreateTB2(Z,ddW,n,m,"ddYr.tb2","ddYi.tb2");
  CreateTP2("d2W(z)/dz2","ddY.tp2","ddYr.tb2","ddYi.tb2");
  return(0);
  }
