/* Traub-Jenkins method for roots of a complex polygon
   developed by Dudley J. Benton
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

double powi(double x,int n)
  {
  double p;
  p=1.;
  while(n>0)
    {
    p*=x;
    n--;
    }
  while(n<0)
    {
    p/=x;
    n++;
    }
  return(p);
  }

double modulus(double cr,double ci)
  {
  double ai,ar;
  ar=fabs(cr);
  ai=fabs(ci);
  if(ar<ai)
    return(ai*sqrt(1.+powi(ar/ai,2)));
  if(ar<=ai)
    return(ar*M_SQRT2);
  return(ar*sqrt(1.+powi(ai/ar,2)));
  }

void ComplexDivide(double ar,double ai,double br,double bi,double*cr,double*ci)
  {
  double d,r;
  if(!(fabs(br)>FLT_MIN||fabs(bi)>FLT_MIN))
    *cr=*ci=FLT_MAX;
  else if(fabs(br)>=fabs(bi))
    {
    r=bi/br;
    d=br+r*bi;
    d=1./d;
    *cr=(ar+ai*r)*d;
    *ci=(ai-ar*r)*d;
    }
  else
    {
    r=br/bi;
    d=bi+r*br;
    d=1./d;
    *cr=(ar*r+ai)*d;
    *ci=(ai*r-ar)*d;
    }
  }

double BoundPoly(int nn,double*qr,double*qi,double rms,double rmp,double are,double rmre)
  {
  int i;
  double e;
  e=modulus(qr[0],qi[0])*rmre/(are+rmre);
  for(i=0;i<nn;i++)
    e=e*rms+modulus(qr[i],qi[i]);
  return(e*(are+rmre)-rmp*rmre);
  }

double LowerBound(int nn,double*pt,double*q)
  {
  int i,n;
  double x;
  double df,dx,f,xm;
  n=nn-1;
  pt[nn-1]=-pt[nn-1];
  x=exp((log(-pt[nn-1])-log(pt[0]))/n);
  if(fabs(pt[n-1])>FLT_MIN)
    {
    xm=-pt[nn-1]/pt[n-1];
    if(xm<x)
      x=xm;
    }
  while(1)
    {
    xm=x*0.1;
    f=pt[0];
    for(i=1;i<nn;i++)
      f=f*xm+pt[i];
    if(f<=0.)
      break;
    x=xm;
    }
  dx=x;
  while(fabs(x)>FLT_MIN)
    {
    if(fabs(dx/x)<=0.005)
      break;
    q[0]=pt[0];
    for(i=1;i<nn;i++)
      q[i]=q[i-1]*x+pt[i];
    f=q[nn-1];
    df=q[0];
    for(i=1;i<n;i++)
      df=df*x+q[i];
    dx=f/df;
    x-=dx;
    }
  return(x);
  }

double ScaleFactor(int nn,double*pt)
  {
  int i,l;
  double rhi,rlo,rmax,rmin,sc,x;
  rhi=sqrt(FLT_MAX);
  rlo=FLT_MIN/DBL_EPSILON;
  rmax=0.;
  rmin=FLT_MAX;
  for(i=0;i<nn;i++)
    {
    x=pt[i];
    if(x>rmax)
      rmax=x;
    if(fabs(x)>FLT_MIN&&x<rmin)
      rmin=x;
    }
  if(!(rmin>=rlo&&rmax<=rhi))
    {
    x=rlo/rmin;
    if(x>1.)
      {
      sc=x;
      if(FLT_MAX/sc<rmax)
        sc=1.;
      }
    else
      sc=1./(sqrt(rmax)*sqrt(rmin));
    l=(int)(log(sc)/log(2.)+0.5);
    return(powi(2.,l));
    }
  return(1.);
  }

void ScalePoly(int nn,double sr,double si,double*pr,double*pi,double*qr,double*qi,double*pvr,double*pvi)
  {
  int i;
  double t;
  qr[0]=pr[0];
  qi[0]=pi[0];
  *pvr=qr[0];
  *pvi=qi[0];
  for(i=1;i<nn;i++)
    {
    t=*pvr*sr-*pvi*si+pr[i];
    *pvi=*pvr*si+*pvi*sr+pi[i];
    *pvr=t;
    qr[i]=*pvr;
    qi[i]=*pvi;
    }
  }

struct{
  int nn;
  double are;
  double*hi;
  double*hr;
  double*pi;
  double*pr;
  double pvi;
  double pvr;
  double*qhi;
  double*qhr;
  double*qpi;
  double*qpr;
  double rmre;
  double*shi;
  double*shr;
  double si;
  double sr;
  double ti;
  double tr;
  }tj;

void dPdH(int l1)
  {
  int i,j,jj,n;
  double t1,t2,xni;
  n=tj.nn-1;
  for(i=0;i<n;i++)
    {
    xni=tj.nn-i-1;
    tj.hr[i]=xni*tj.pr[i]/n;
    tj.hi[i]=xni*tj.pi[i]/n;
    }
  for(jj=1;jj<=l1;jj++)
    {
    if(modulus(tj.hr[n-1],tj.hi[n-1])<=DBL_EPSILON*10.*modulus(tj.pr[n-1],tj.pi[n-1]))
      {
      for(i=1;i<=n-1;i++)
        {
        j=tj.nn-i;
        tj.hr[j-1]=tj.hr[j-2];
        tj.hi[j-1]=tj.hi[j-2];
        }
      tj.hr[0]=tj.hi[0]=0.;
      }
    else
      {
      ComplexDivide(-tj.pr[tj.nn-1],-tj.pi[tj.nn-1],tj.hr[n-1],tj.hi[n-1],&tj.tr,&tj.ti);
      for(i=1;i<=n-1;i++)
        {
        j=tj.nn-i;
        t1=tj.hr[j-2];
        t2=tj.hi[j-2];
        tj.hr[j-1]=tj.tr*t1-tj.ti*t2+tj.pr[j-1];
        tj.hi[j-1]=tj.tr*t2+tj.ti*t1+tj.pi[j-1];
        }
      tj.hr[0]=tj.pr[0];
      tj.hi[0]=tj.pi[0];
      }
    }
  }

BOOL BowlTest()
  {
  BOOL bowl;
  int n;
  double hvi,hvr;
  n=tj.nn-1;
  ScalePoly(n,tj.sr,tj.si,tj.hr,tj.hi,tj.qhr,tj.qhi,&hvr,&hvi);
  bowl=modulus(hvr,hvi)<=tj.are*10.*modulus(tj.hr[n-1],tj.hi[n-1]);
  if(bowl)
    tj.tr=tj.ti=0.;
  else
    ComplexDivide(-tj.pvr,-tj.pvi,hvr,hvi,&tj.tr,&tj.ti);
  return(bowl);
  }

void ShiftPoly(BOOL bowl)
  {
  int j,n;
  double t1,t2;
  n=tj.nn-1;
  if(bowl)
    {
    for(j=1;j<n;j++)
      {
      tj.hr[j]=tj.qhr[j-1];
      tj.hi[j]=tj.qhi[j-1];
      }
    tj.hr[0]=tj.hi[0]=0.;
    }
  else
    {
    for(j=1;j<n;j++)
      {
      t1=tj.qhr[j-1];
      t2=tj.qhi[j-1];
      tj.hr[j]=tj.tr*t1-tj.ti*t2+tj.qpr[j];
      tj.hi[j]=tj.tr*t2+tj.ti*t1+tj.qpi[j];
      }
    tj.hr[0]=tj.qpr[0];
    tj.hi[0]=tj.qpi[0];
    }
  }

void PolyZero1(int l3,double*zr,double*zi,BOOL*conv)
  {
  BOOL b,bowl;
  int i,j;
  double omp,r1,r2,relstp,rmp,rms,tp;
  *conv=FALSE;
  b=FALSE;
  tj.sr=*zr;
  tj.si=*zi;
  for(i=1;i<=l3;i++)
    {
    ScalePoly(tj.nn,tj.sr,tj.si,tj.pr,tj.pi,tj.qpr,tj.qpi,&tj.pvr,&tj.pvi);
    rmp=modulus(tj.pvr,tj.pvi);
    rms=modulus(tj.sr,tj.si);
    if(rmp<=20.*BoundPoly(tj.nn,tj.qpr,tj.qpi,rms,rmp,tj.are,tj.rmre))
      {
      *conv=TRUE;
      *zr=tj.sr;
      *zi=tj.si;
      return;
      }
    if(i!=1)
      {
      if(!((b||rmp<omp)||relstp>=0.05))
        {
        tp=relstp;
        b=TRUE;
        if(relstp<DBL_EPSILON)
          tp=DBL_EPSILON;
        r1=sqrt(tp);
        r2=tj.sr*(1.+r1)-tj.si*r1;
        tj.si=tj.sr*r1+tj.si*(1.+r1);
        tj.sr=r2;
        ScalePoly(tj.nn,tj.sr,tj.si,tj.pr,tj.pi,tj.qpr,tj.qpi,&tj.pvr,&tj.pvi);
        for(j=0;j<5;j++)
          {
          bowl=BowlTest();
          ShiftPoly(bowl);
          }
        omp=FLT_MAX;
        goto next;
        }
      else if(rmp*0.1>omp)
        break;
      }
    omp=rmp;
  next:
    bowl=BowlTest();
    ShiftPoly(bowl);
    bowl=BowlTest();
    if(!bowl)
      {
      relstp=modulus(tj.tr,tj.ti)/modulus(tj.sr,tj.si);
      tj.sr+=tj.tr;
      tj.si+=tj.ti;
      }
    }
  }

void PolyZero2(int l2,double*zr,double*zi,BOOL*conv)
  {
  BOOL bowl,pasd,test;
  int i,j,n;
  double oti,otr,svsi,svsr;
  n=tj.nn-1;
  ScalePoly(tj.nn,tj.sr,tj.si,tj.pr,tj.pi,tj.qpr,tj.qpi,&tj.pvr,&tj.pvi);
  test=TRUE;
  pasd=FALSE;
  bowl=BowlTest();
  for(j=1;j<=l2;j++)
    {
    otr=tj.tr;
    oti=tj.ti;
    ShiftPoly(bowl);
    bowl=BowlTest();
    *zr=tj.sr+tj.tr;
    *zi=tj.si+tj.ti;
    if(!((bowl||!test)||j==l2))
      {
      if(modulus(tj.tr-otr,tj.ti-oti)>=0.5*modulus(*zr,*zi))
        pasd=FALSE;
      else if(pasd)
        {
        for(i=0;i<n;i++)
          {
          tj.shr[i]=tj.hr[i];
          tj.shi[i]=tj.hi[i];
          }
        svsr=tj.sr;
        svsi=tj.si;
        PolyZero1(10,zr,zi,conv);
        if(*conv)
          return;
        test=FALSE;
        for(i=0;i<n;i++)
          {
          tj.hr[i]=tj.shr[i];
          tj.hi[i]=tj.shi[i];
          }
        tj.sr=svsr;
        tj.si=svsi;
        ScalePoly(tj.nn,tj.sr,tj.si,tj.pr,tj.pi,tj.qpr,tj.qpi,&tj.pvr,&tj.pvi);
        bowl=BowlTest();
        }
      else
        pasd=TRUE;
      }
    }
  PolyZero1(10,zr,zi,conv);
  }

int TraubJenkins(double*a,int ndeg,double*z)
  {
  BOOL conv;
  int i,icnt1,icnt2,ii,inx,inxi,j,n2,nn2;
  double bnd,cosr,sinr,xx,xxx,yy,zi,zr;
  if(ndeg<1)
    return(__LINE__);
  if((tj.hi=calloc((ndeg+1)*10,sizeof(double)))==NULL)
    return(__LINE__);
  tj.hr=tj.hi+ndeg+1;
  tj.pi=tj.hr+ndeg+1;
  tj.pr=tj.pi+ndeg+1;
  tj.qhi=tj.pr+ndeg+1;
  tj.qhr=tj.qhi+ndeg+1;
  tj.qpi=tj.qhr+ndeg+1;
  tj.qpr=tj.qpi+ndeg+1;
  tj.shi=tj.qpr+ndeg+1;
  tj.shr=tj.shi+ndeg+1;
  tj.are=DBL_EPSILON;
  tj.rmre=2.*M_SQRT2*DBL_EPSILON;
  xx=M_SQRT1_2;
  yy=-xx;
  cosr=-0.06975647;
  sinr=0.9975641;
  tj.nn=ndeg+1;
  if(fabs(a[0])>FLT_MIN||fabs(a[1])>FLT_MIN)
    {
    while(1)
      {
      nn2=tj.nn+tj.nn;
      if(fabs(a[nn2-2])>FLT_MIN||fabs(a[nn2-1])>FLT_MIN)
        break;
      inx=ndeg-tj.nn+2;
      inxi=inx+ndeg;
      z[inxi-1]=z[inx-1]=0.;
      tj.nn--;
      if(tj.nn==1)
        {
        free(tj.hi);
        return(0);
        }
      }
    for(i=1;i<=tj.nn;i++)
      {
      ii=i+i;
      tj.pr[i-1]=a[ii-2];
      tj.pi[i-1]=a[ii-1];
      tj.shr[i-1]=modulus(tj.pr[i-1],tj.pi[i-1]);
      }
    bnd=ScaleFactor(tj.nn,tj.shr);
    if(bnd!=1.)
      {
      for(i=0;i<tj.nn;i++)
        {
        tj.pr[i]*=bnd;
        tj.pi[i]*=bnd;
        }
      }
    while(tj.nn>2)
      {
      for(i=0;i<tj.nn;i++)
        tj.shr[i]=modulus(tj.pr[i],tj.pi[i]);
      bnd=LowerBound(tj.nn,tj.shr,tj.shi);
      for(icnt1=1;icnt1<=2;icnt1++)
        {
        dPdH(5);
        for(icnt2=1;icnt2<=9;icnt2++)
          {
          xxx=cosr*xx-sinr*yy;
          yy=sinr*xx+cosr*yy;
          xx=xxx;
          tj.sr=bnd*xx;
          tj.si=bnd*yy;
          PolyZero2(10*icnt2,&zr,&zi,&conv);
          if(conv)
            goto next;
          }
        }
      free(tj.hi);
      return(__LINE__);
  next:
      inx=ndeg+2-tj.nn;
      inxi=inx+ndeg;
      z[inx-1]=zr;
      z[inxi-1]=zi;
      tj.nn--;
      for(i=0;i<tj.nn;i++)
        {
        tj.pr[i]=tj.qpr[i];
        tj.pi[i]=tj.qpi[i];
        }
      }
    ComplexDivide(-tj.pr[1],-tj.pi[1],tj.pr[0],tj.pi[0],&z[ndeg-1],&z[ndeg+ndeg-1]);
    for(i=1;i<=ndeg;i++)
      tj.pi[i-1]=z[ndeg+i-1];
    n2=ndeg+ndeg;
    j=ndeg;
    for(i=1;i<=ndeg;i++)
      {
      z[n2-2]=z[j-1];
      z[n2-1]=tj.pi[j-1];
      n2-=2;
      j--;
      }
    free(tj.hi);
    return(0);
    }
  free(tj.hi);
  return(__LINE__);
  }

int randbetween(int lo,int hi)
  {
  return(lo+rand()%(hi-lo+1));
  }

int main(int argc,char**argv,char**envp)
  {
  int i,ier,j;
  double a[8]={1.,2.,3.,4.,5.,6.,7.,8.},z[6];
  for(j=0;j<10;j++)
    {
    printf("a");
    for(i=0;i<sizeof(a)/sizeof(double);i+=2)
      printf("%c(%lG%+lGi)",i?',':'=',a[i],a[i+1]);
    printf("\n");
    if((ier=TraubJenkins(a,3,z))!=0)
      printf("ier=%i\n",ier);
    else
      {
      printf("z");
      for(i=0;i<sizeof(z)/sizeof(double);i+=2)
        printf("%c(%lG%+lGi)",i?',':'=',z[i],z[i+1]);
      printf("\n");
      }
    for(i=0;i<sizeof(a)/sizeof(double);i++)
      a[i]=randbetween(-10000,10000)/1000.;
    }
  return(0);
  }
