/* Gauss elimination with row and column pivoting
   plus linear least-squares regression
   by Dudley J. Benton
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "../include/complex.hpp"

int SimultaneousEquations(complex*A,complex*B,int n)
  {
  int i,j,k,ip,jp,*pivot;
  double p,q;
  complex a,b;
  if(n<1)
    return(__LINE__);
  if(n==1)
    {
    if(abs(A[0])<FLT_MIN)
      return(__LINE__);
    B[0]/=A[0];
    return(0);
    }
  if((pivot=(int*)calloc(n,sizeof(int)))==NULL)
    return(__LINE__);
  for(i=0;i<n;i++)
    pivot[i]=i;
  for(k=0;k<n-1;k++)
    {
    ip=k;
    jp=k;
    p=abs(A[k*(n+1)]);
    for(i=k;i<n;i++)
      {
      for(j=k;j<n;j++)
        {
        q=abs(A[n*i+j]);
        if(q>p)
          {
          ip=i;
          jp=j;
          p=q;
          }
        }
      }
    if(p<FLT_MIN)
      {
      free(pivot);
      return(__LINE__);
      }
    if(ip!=k)
      {
      b=B[k];
      B[k]=B[ip];
      B[ip]=b;
      for(j=k;j<n;j++)
        {
        a=A[n*k+j];
        A[n*k+j]=A[n*ip+j];
        A[n*ip+j]=a;
        }
      }
    if(jp!=k)
      {
      j=pivot[jp];
      pivot[jp]=pivot[k];
      pivot[k]=j;
      for(i=0;i<n;i++)
        {
        a=A[n*i+k];
        A[n*i+k]=A[n*i+jp];
        A[n*i+jp]=a;
        }
      }
    for(i=k+1;i<n;i++)
      {
      a=A[n*i+k]/A[n*k+k];
      B[i]-=a*B[k];
      for(j=k+1;j<n;j++)
        A[n*i+j]-=a*A[n*k+j];
      }
    }
  a=A[n*n-1];
  p=abs(a);
  if(p<FLT_MIN)
    {
    free(pivot);
    return(__LINE__);
    }
  B[n-1]/=a;
  for(k=1;k<n;k++)
    {
    i=n-1-k;
    b=0.;
    for(j=i+1;j<n;j++)
      b+=A[n*i+j]*B[j];
    B[i]=(B[i]-b)/A[n*i+i];
    }
  for(i=0;i<n;i++)
    A[i]=B[i];
  for(i=0;i<n;i++)
    B[pivot[i]]=A[i];
  free(pivot);
  return(0);
  }

int LinearLeastSquares(complex*x,complex*y,int m,complex*a,int n)
  {
  int i,j,k;
  complex*ata,*atb;
  if(n<2)
    return(__LINE__);
  if(m<n)
    return(__LINE__);
  if((ata=(complex*)calloc(n*(n+1),sizeof(complex)))==NULL)
    return(__LINE__);
  atb=ata+n*n;
  for(k=0;k<m;k++)
    {
    a[0]=1.;
    for(i=1;i<n;i++)
      a[i]=a[i-1]*x[k];
    for(i=0;i<n;i++)
      {
      atb[i]+=a[i]*y[k];
      for(j=0;j<n;j++)
        ata[n*i+j]+=a[i]*a[j];
      }
    }
  i=SimultaneousEquations(ata,atb,n);
  for(j=0;j<n;j++)
    a[j]=atb[j];
  free(ata);
  if(i)
    return(i);
  return(0);
  }

complex horner(complex*c,int m,complex z)
  {
  int i;
  static complex w;
  for(w.re=w.im=i=0;i<m;i++)
    {
    w*=z;
    w+=c[m-i-1];
    }
  return(w);
  }

int randbetween(int lo,int hi)
  {
  return(lo+rand()%(hi-lo+1));
  }

#define n   9 /* number of terms */
#define m 120 /* number of points */

int main(int argc,char**argv,char**envp)
  {
  int i;
  static complex c[n],f[m],w,z[m];
  printf("calculating %i values at random points\n",m);
  for(i=0;i<m;i++)
    {
    z[i]=complex(randbetween(-2000,2000)/1000.,randbetween(-2000,2000)/1000.);
    f[i]=cosh(z[i]);
    }
  printf("solving linear least-squares problem\n");
  i=LinearLeastSquares(z,f,m,c,n);
  printf("error=%i\n",i);
  if(i)
    return(0);
  printf("c[%i]",n);
  for(i=0;i<n;i++)
    printf("%c(%+lG%+lGi)",i?',':'=',c[i].re,c[i].im);
  printf("\n");
  printf("     random points    target value    regression\n");
  for(i=0;i<m;i++)
    {
    w=horner(c,n,z[i]);
    printf("cosh(%+6.3lf%+6.3lfi)=(%+6.3lf%+6.3lfi)?(%+6.3lf%+6.3lfi)\n",z[i].re,z[i].im,f[i].re,f[i].im,w.re,w.im);
    }
  return(0);
  }
